//----------------------------------------------------------------
// Lighting class
//----------------------------------------------------------------
#ifndef __OP_LIGHT_H
#define __OP_LIGHT_H
#ifndef __OP_TYPES_H
#include <op_types.h>
#endif
#ifndef __OP_VECT_H
#include <op_vect.h>
#endif

//--------------------------------------------------------------------------
// Lighting class - double version
//--------------------------------------------------------------------------
class d_Light
{
 private:
  Vect *vect;            // Private Vector geometry class
  d_source *source;      // Array of light sources
  int max_lights;        // Maximum number of light sources definable
  int light_scale;       // Current light drop-off scaling factor

 public:
  d_Light(int l_max = 1, int l_scale = LIGHTSCALE);
  ~d_Light();
  int changed;           // Set when the light source is moved
  void define_source(int spot, double x = 0, double y = 0, double z = 0,
                     double r = 1.0, double g = 1.0, double b = 1.0);
  void chg_pos(int spot, double x_move, double y_move, double z_move);
  void chg_color(int spot, double r_chg, double g_chg, double b_chg);
  void determine(d_surface *surface);
};

//--------------------------------------------------------------------------
// Lighting class - fixed point version
//--------------------------------------------------------------------------
class f_Light
{
 private:
  Vect *vect;            // Private Vector geometry class
  f_source *source;      // Array of light sources
  int max_lights;        // Maximum number of light sources definable
  fixed light_scale;     // Current light drop-off scaling factor

 public:
  f_Light(int l_max = 1, fixed l_scale = I2F(LIGHTSCALE));
  ~f_Light();
  int changed;           // Set when the light source is moved
  void define_source(int spot, fixed x = 0, fixed y = 0, fixed z = 0,
                     fixed r = 0, fixed g = 0, fixed b = 0);
  void chg_pos(int spot, fixed x_move, fixed y_move, fixed z_move);
  void chg_color(int spot, fixed r_chg, fixed g_chg, fixed b_chg);
  void determine(f_surface *surface);
};
#endif
