/*-------------------------------------------------------------------------
PCX.H - Part of FAST Viewer project
Description: This file contains prototypes related to handling PCX images.
At present, only viewing of such images and reading pcx header is supported.
The rountine(s) are defined in pcx.cpp
-----------------------------------------------------------------------------
Accessible functions:
int show_pcx (char *name);
void read_pcx_header (FILE *dataFile, pcxHeaderType &pcxHeader);
---------------------------------------------------------------------------*/


/*-------------------------------------------------------------------------
ACCESSIBLE VARIABLES / DATA TYPES DEFINED IN PCX.CPP
--------------------------------------------------------------------------*/
struct pcxHeaderType		// also defined in pcx.cpp
   {
   char manufacturer;		// always 10
   char version;                // usually 5
   char encoding;		// always 1, indicates RLE
   char bitsPerPixel;		// number of bits used to represent a pixel
   int  x, y;			// upper left coordinates
   int  width, height;		// width and height of image
   int  hResolution, vResolution;	// scanned resolutions
   char egaPalette[48];		// dummy for 256 color images
   char reserved;               // unused, always 0
   char nColorPlanes;		// number of color planes
   int  bytesPerLine;		// bytes per scan line
   int  paletteType;		// colored or grey scale
   char padding[58];		// reserved, always 0
   };


/*-------------------------------------------------------------------------
ERROR CODES FOR CALLING PROCEDURE
--------------------------------------------------------------------------*/
char *pcxError[6] = {
	"Successful",
	"Unsupported PCX: Bits Per Pixel not equal to 8",
	"Unsupported PCX: Number of Planes not equal to 1",
	"Invalid PCX",
	"Can not open this file"};


/*-------------------------------------------------------------------------
ACCESSIBLE PROTO TYPES
--------------------------------------------------------------------------*/
int show_pcx (char *name);
void read_pcx_header (FILE *filename, pcxHeaderType &pcxHeader);
/*------------------------------------------------------------------------*/