/*---------------------------------------------------------------------------
MODE13H.CPP - Part of FAST Viewer project
Description: This file contains general 320x200x256 graphics mode routines.
It may be used in some other project as it is an independent module. Just
include "mode13h.h" in that project.
-----------------------------------------------------------------------------
Accessible variables:
. typedef unsigned char byte;	// defined data type - byte
. char far *video		// pointer to video memory 0xA0000000
-----------------------------------------------------------------------------
Accessible functions:
. void set_mcga (void);		// set graphics mode
. void set_text (void);		// set text mode
. void put_pixel (int x, int y, byte color);	// put pixel on screen
. void set_palette (int color, byte r, byte g, byte b);
---------------------------------------------------------------------------*/

#include <dos.h>


/*---------------------------------------------------------------------------
GLOBAL VARIABLES
---------------------------------------------------------------------------*/
typedef unsigned char byte;			// a shorter name
char far *video = (char far *) 0xA0000000;	// pointer to video memory




/*---------------------------------------------------------------------------
FUNCTION:	Sets graphics mode 13h
INPUT:		None
PROCESS:	Uses interrupts to set 320x200 resolution with 256 colors
OUTPUT:		None
---------------------------------------------------------------------------*/
void set_mcga (void)
   {
   union REGS regs;		// defined in dos.h - registers
   regs.h.ah = 0;		// service number to change video mode
   regs.h.al = 0x13;		// mode 13h
   int86 (0x10, &regs, &regs);
   }
/*-------------------------------------------------------------------------*/




/*---------------------------------------------------------------------------
FUNCTION:	Sets text mode
INPUT:		None
PROCESS:	Uses interrupts to set normal text mode
OUTPUT:		None
---------------------------------------------------------------------------*/
void set_text (void)
   {
   union REGS regs;		// defined in dos.h - registers
   regs.h.ah = 0;		// to change video mode
   regs.h.al = 0x03;		// text mode number
   int86 (0x10, &regs, &regs);
   }
/*-------------------------------------------------------------------------*/




/*---------------------------------------------------------------------------
FUNCTION:	Sets a color in palette
INPUT:		Color number to set and values of RGB
PROCESS:	Uses port acess to set a color in palette
OUTPUT:		None
---------------------------------------------------------------------------*/
void set_palette (int color, byte r, byte g, byte b)
   {
   outportb (0x3C8, color);	// write color number to change at 0x3C8
   outportb (0x3C9, r);		// write values for red, green and blue
   outportb (0x3C9, g);
   outportb (0x3C9, b);
   }
/*-------------------------------------------------------------------------*/




/*---------------------------------------------------------------------------
FUNCTION:	Puts a pixel of given color at x, y
INPUT:		Color number to set and coordinates
PROCESS:	Uses *video to put individual pixels
OUTPUT:		None
---------------------------------------------------------------------------*/
void put_pixel (int x, int y, byte color)
   {
   if (y>=0 && y<200)			// can we plot this pixel?
      if (x>=0 && x<320)
	 video[y*320+x] = color;	// use direct memory access
   }
/*-------------------------------------------------------------------------*/
