/*---------------------------------------------------------------------------
PCX.CPP - Part of FAST Viewer project
Description: This file handles pcx format. At present, there are only two
accessible function: one that displays pcx files on screen and the other
reads header of a pcx file. This module  may be used in some other project
as it is an independent module. Just include "pcx.h" in that project.
-----------------------------------------------------------------------------
Accessible functions:
int show_pcx (char *name);
void read_pcx_header (FILE *dataFile, pcxHeaderType &pcxHeader);
---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include "grafx2.cpp"


/*---------------------------------------------------------------------------
GLOBAL VARIABLES
---------------------------------------------------------------------------*/
struct pcxHeaderType		// also defined in pcx.h
   {
   char manufacturer;		// always 10
   char version;                // usually 5
   char encoding;		// always 1, indicates RLE
   char bitsPerPixel;		// number of bits used to represent a pixel
   int  x, y;			// upper left coordinates
   int  width, height;		// width and height of image
   int  hResolution, vResolution;	// scanned resolutions
   char egaPalette[48];		// dummy for 256 color images
   char reserved;               // unused, always 0
   char nColorPlanes;		// number of color planes
   int  bytesPerLine;		// bytes per scan line
   int  paletteType;		// colored or grey scale
   char padding[58];		// reserved, always 0
   };




/*---------------------------------------------------------------------------
FUNCTION:	Reads header of a pcx file
INPUT:		Name of file and header reference variable
PROCESS:	Opens up the file and reads in its header using fread
OUTPUT:		None
---------------------------------------------------------------------------*/
void read_pcx_header (FILE *dataFile, pcxHeaderType &pcxHeader)
   {
   fread (&pcxHeader, sizeof (pcxHeaderType), 1, dataFile);
   }
/*-------------------------------------------------------------------------*/

typedef unsigned char byte;

/*---------------------------------------------------------------------------
FUNCTION:	Shows a pcx file on screen
INPUT:		Name of file to be shown
PROCESS:	Opens up the file, interpret its header and shows it on the
		screen. Uses routines defined in mode13h.cpp.
OUTPUT:		Error code
---------------------------------------------------------------------------*/
int show_pcx (char *name)
{

   pcxHeaderType pcxHeader;	// structure to store header
   FILE *dataFile; 		// file pointer to pcx image
   byte color, r, g, b,		// for palette settings
	counter, length;	// for RLE
   unsigned int x, y;		// coordinates of pixel to plot

   dataFile = fopen (name, "rb");	// try to open the image
   if (dataFile == NULL) return 4;	// could not open the file?

   read_pcx_header (dataFile, pcxHeader);	// and check for errors
   if (pcxHeader.bitsPerPixel != 8)	// unsupported pcx
      { fclose (dataFile); return 1; }
   if (pcxHeader.nColorPlanes != 1)  	// unsupported pcx
      { fclose (dataFile); return 2; }
   if (pcxHeader.manufacturer != 10)    // invalid pcx
      { fclose (dataFile); return 3; }


   fseek (dataFile, -768, SEEK_END);	// start reading the image
   SetVGA();				// set graphics mode
   for (x=0; x<256; x++)		// read palette
     {
     r = fgetc (dataFile); r = r>>2;	// stored as rgb
     g = fgetc (dataFile); g = g>>2;
     b = fgetc (dataFile); b = b>>2;
     SetPal (x, r, g, b);
     }

   // start decoding data
   x=0; y=0;				// upper left coordinates
   fseek (dataFile, 128, SEEK_SET);	// data offset
   while (y <= pcxHeader.height)	// until whole image done
      {
      color = fgetc (dataFile);		// read a byte
      if (color < 192)			// uncompressed?
	 {
	 if (x > pcxHeader.width) { x=0; y++; }	// check for overflow
	 Putpixel (x++, y, color,vga);		// plot it
	 }

      else				// must be compressed
	 {
	 length = color-192;		// number of times next byte repeated
	 color = fgetc (dataFile);	// read next byte
	 for (counter=0; counter<length; counter++)
	    {
	    if (x > pcxHeader.width) { x=0; y++; }	// overflow?
	    Putpixel (x++, y, color,vga);			// plot it
	    }
	 }
      }

   fclose (dataFile);	// close data file
   getch();		// wait for a key press
   return 0;		// reached here? then no error occured
   }
/*-------------------------------------------------------------------------*/