/* Demonstracao do uso da linguagem C para o tracado de graficos.
   Tracado da curva da funcao cossenoidal modulada em amplitude (AM)
   y=(a*cos(b*3.1415*x/360)*c*sin(b*3.1415*x*50/360)),
   com possibilidade de alteracao dos valores de a, b e c e efeitos
   de deslocamento para a esquerda/direita e para cima/baixo (pan)
   e reducao/ampliacao (zoom in/out) da curva durante a visualizacao,
   a partir do uso das teclas:

   a - aumento da amplitude
   r - reducao da amplitude
   e - deslocamento do eixo para a esquerda
   d - deslocamento do eixo para a direita
   c - deslocamento do eixo para cima
   b - deslocamento do eixo para baixo

   - - reducao no eixo das ordenadas
   + - ampliacao no eixo das ordenadas
   [ - reducao no eixo das abcissas
   ] - ampliacao no eixo das abcissas

   / - Zoom default (inicial)

   Importante!!! A execucao deste programa exige que o arquivo EGAVGA.BGI se
   encontre no mesmo diretorio do arquivo executavel */

#include <graphics.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <conio.h>
#include <math.h>

/* Possibilidade de alteracao de valores da resolucao de tela */

#define MAXX 640
#define MAXY 480

/* Prototipos e declaracoes de funcoes de conversao de coordenadas */

float tela_y(float, float, float);
float tela_y(float, float, float);

/* Possibilidade de definicao da funcao a ser tracada */

float matfunc(float);

/* Possibilidade de uso de todos os parametros a seguir como coeficientes
   para qualquer funcao cujo tracado se deseja. Quaisquer atribuicoes em
   main() e definicao da funcao em matfunc() */

float a,b,c,d,e,f,g,h,i,j;

float tela_x(float xb, float x, float xe)
{
	return ((x-xb)/(xe-xb)*MAXX);
}

float tela_y(float yb, float y, float ye)
{
	return (MAXY-(y-yb)/(ye-yb)*MAXY);
}

float matfunc(float x)
{
	/* Definicao da funcao cujo tracado se deseja */

	/* return a * cos(( b * x + c ) * 3.1415 / 180.) ; */
	return (a*cos(b*3.1415*x/360)*c*sin(b*3.1415*x*50/360));
	//return (0.25*a*(exp(-0.009*x))*cos(b*3.1415*x*10/360));
}

/* Armazenamento das instrucoes de visualizacao do grafico */
char  *menu[] = { "a - AUMENTO da amplitude",
		  "r - REDUCAO da amplitude",
		  "e - deslocamento do eixo p/ a ESQUERDA",
		  "d - deslocamento do eixo p/ a DIREITA",
		  "c - deslocamento do eixo p/ CIMA",
		  "b - deslocamento do eixo p/ BAIXO",
		  "- - REDUCAO no eixo das ordenadas",
		  "+ - AMPLIACAO no eixo das ordenadas",
		  "[ - REDUCAO no eixo das abcissas",
		  "] - AMPLIACAO no eixo das abcissas",
		  "/ - ZOOM default (inicial)"};

void main(void)
{
	/* Possibilidade de alteracao dos valores de 'drv' e 'modo' em funcao
	da alteracao de MAXX e MAXY */

	int m,drv=VGA, modo=VGAHI;
	float xb,xe,yb,ye,x,y,potx,poty;
	char tecla;

	char estilostr[50], estilostr1[50];

	/* Parametros de contexto configuraveis */
	a  = 1;
	b  = 1;
	c  = 2;
	xb = -360;
	xe = 360;
	yb = -10;
	ye = 10;

	/* Inicializacao do modo grafico */


	initgraph(&drv,&modo,"");

	/* Tracado dos eixos X e Y */
	setcolor(1);
	line(0,tela_y(yb,0,ye),MAXX,tela_y(yb,0,ye));
	line(tela_x(xb,0,xe),0,tela_x(xb,0,xe),MAXY);

	setcolor(15);
	for(m=0;m<=10;m++)
	{
		/* Conversao das opcoes de menu em strings */
		strcpy(estilostr, menu[m]);

		/* Apresentacao do menu */
		outtextxy(330, 20 + 12*m, estilostr);
	}

	setcolor(14);
	moveto(0,tela_y(yb,matfunc(xb),ye));
	for(x=xb;x<=xe;x+=(xe-xb)/MAXX)
	{
		y = matfunc(x);
		lineto(tela_x(xb,x,xe),tela_y(yb,y,ye));
	}

	while((tecla=getch())!=27) /* Repeticao do laco ate ativacao de ESC */
	{
		potx = xb;
		poty = yb;
		switch(tecla)
		{
			case 'a':   /* Aumento da amplitude */
				++a;
				break;
			case 'r':   /* Reducao da amplitude */
				--a;
				break;
			case 'd':   /* Deslocamento do eixo para a direita */
				xb -= (xe-xb)/4.0;
				xe -= (xe-potx)/4.0;
				break;
			case 'e':   /* Deslocamento do eixo para a esquerda */
				xb += (xe-xb)/4.0;
				xe += (xe-potx)/4.0;
				break;
			case 'b':   /* Deslocamento do eixo para baixo */
				yb += (ye-yb)/4.0;
				ye += (ye-poty)/4.0;
				break;
			case 'c':   /* Deslocamento do eixo para cima */
				yb -= (ye-yb)/4.0;
				ye -= (ye-poty)/4.0;
				break;
			case '+':   /* Ampliacao do eixo Y */
				yb -= (yb-ye)/4.0;
				ye += (poty-ye)/4.0;
				break;
			case '-':   /* Reducao do eixo Y */
				yb += (yb-ye)/4.0;
				ye -= (poty-ye)/4.0;
				break;
			case '[':   /* Reducao do eixo X */
				xb += (xb-xe)/4.0;
				xe -= (potx-xe)/4.0;
				break;
			case ']':   /* Ampliacao do eixo X */
				xb -= (xb-xe)/4.0;
				xe += (potx-xe)/4.0;
				break;
			case '/':   /* Visualizacao inicial */
				xb = -360;
				xe = 360;
				yb = 10;
				ye = -10;
				break;
			default:
				continue;
				//break;
		}

		cleardevice();

		setcolor(1);
		line(0,tela_y(yb,0,ye),MAXX,tela_y(yb,0,ye));
		line(tela_x(xb,0,xe),0,tela_x(xb,0,xe),MAXY);
		if((tecla=='e')||(tecla=='[')||(tecla==']')||(tecla=='/')||(tecla=='r'))
		{
			setcolor(15);
			for(m=0;m<=8;m++)
			{
				/* Conversao das opcoes de menu em strings */
				strcpy(estilostr, menu[m]);

				/* Apresentacao do menu */
				outtextxy(330, 20 + 12*m, estilostr);
			}
		}
		else if(tecla=='d')
		{
			setcolor(15);
			for(m=0;m<=10;m++)
			{
				/* Conversao das opcoes de menu em strings */
				strcpy(estilostr, menu[m]);

				/* Apresentacao do menu */
				outtextxy(30, 20 + 12*m, estilostr);
			}
		}
		else if(tecla=='b')
		{
			setcolor(15);
			for(m=0;m<=10;m++)
			{
				/* Conversao das opcoes de menu em strings */
				strcpy(estilostr, menu[m]);

				/* Apresentacao do menu */
				outtextxy(330, 20 + 12*m, estilostr);
			}
		}
		else if(tecla=='c')
		{
			setcolor(15);
			for(m=0;m<=10;m++)
			{
				/* Conversao das opcoes de menu em strings */
				strcpy(estilostr, menu[m]);

				/* Apresentacao do menu */
				outtextxy(330, 290 + 12*m, estilostr);
			}
		}
		setcolor(14);
		moveto(0,tela_y(yb,matfunc(xb),ye));
		for(x=xb;x<=xe;x+=(xe-xb)/MAXX)
		{
			y = matfunc(x);
			lineto(tela_x(xb,x,xe),tela_y(yb,y,ye));
		}

		setcolor(14);
		moveto(0,tela_y(yb,matfunc(xb),ye));
		for(x=xb;x<=xe;x+=(xe-xb)/MAXX)
		{
			y = matfunc(x);
			lineto(tela_x(xb,x,xe),tela_y(yb,y,ye));
		}

	}
	closegraph();
	return;
}